@@ -13,6 +13,13 @@ MESSAGES = {
         'Used when an undefined field of a protobuf generated class is '
         'accessed'
     ),
+    'E%02d02' % BASE_ID: (
+        'Value %r does not appear in the declared values of protobuf-'
+        'generated enum %r and will raise ValueError at runtime',
+        'protobuf-enum-value',
+        'Used when an undefined enum value of a protobuf generated enum '
+        'is built'
+    ),
 }
 PROTOBUF_IMPLICIT_ATTRS = [
     'ByteSize',
@@ -37,6 +44,7 @@ PROTOBUF_IMPLICIT_ATTRS = [
 
 
 class TypeTags(object):
+    NONE = -1
     OBJECT = 11
 
 
@@ -163,7 +171,10 @@ def _typeof(scope, node):
     else:
         if node is None:
             return None  # node may be Uninferable
-        return scope.get(node.as_string())
+        try:
+            return scope.get(node.as_string())
+        except AttributeError:
+            return None
 
 
 def _assign(scope, target, rhs):
@@ -194,12 +205,15 @@ def _assignattr(scope, node):
     expr_type = _typeof(scope, expr)
     if expr_type is None or isinstance(expr_type, Module):
         return True, []  # not something we're tracking?
-    fields = expr_type.fields  # ClassDef
+    try:
+        fields, qualname = expr_type.fields, expr_type.qualname  # ClassDef
+    except AttributeError:
+        fields, qualname = expr_type.t.fields, expr_type.t.qualname  # TypeClass
     if fields is None:
         # assert False, "type fields missing for {!r}".format(expr_type)
         return False, []
     if attr not in fields and attr not in PROTOBUF_IMPLICIT_ATTRS:
-        msg = ('protobuf-undefined-attribute', (attr, expr_type.qualname), node)
+        msg = ('protobuf-undefined-attribute', (attr, qualname), node)
         return False, [msg]
     return False, []
 
@@ -277,6 +291,8 @@ def _parse_fields(iterable, inner_fields, qualname):
                 continue
             fully_qualified_name = '{}.{}'.format(qualname, name)
             fields[name] = ClassDef(desc, fully_qualified_name)
+        elif type_ == TypeTags.NONE:  # EnumValueDescriptor?
+            fields[name] = SimpleField(name)
         else:
             fields[name] = SimpleField(name)  # FIXME
     return fields
@@ -301,6 +317,28 @@ def _parse_descriptor(node, candidates, qualname):
     return None
 
 
+def _parse_enum_descriptor(node, candidates, qualname):
+    # TODO: docstring, unify with above
+    """
+    Walk the nodes of a descriptor_pb2.Descriptor to find the fields keyword.
+    """
+    if node is None:
+        return None
+    node = node[0]
+    assignment = node.parent
+    if not isinstance(assignment, astroid.Assign):
+        return None
+    call = assignment.value
+    if not isinstance(call, astroid.Call):
+        return None
+    for kw in call.keywords:
+        if kw.arg == 'values':
+            fields = _parse_fields(kw.value.itered(), candidates, qualname)
+            fields['Value'] = SimpleField('Value')  # implicit
+            return fields
+    return None
+
+
 def _extract_fields(node, module_globals, inner_fields, qualname):
     """
     Given a "name = type(...)"-style assignment, look up the variable
@@ -321,15 +359,33 @@ def _extract_fields(node, module_globals, inner_fields, qualname):
     if inner_fields is None:
         inner_fields = {}
 
+    if not isinstance(node, astroid.AssignName):
+        return None
+    rhs = node.parent.value
+    if isinstance(rhs, astroid.Const):
+        return _parse_enum_value(node, rhs)
+    try:
+        attr_name = rhs.func.attrname
+    except AttributeError:
+        return None
+    if attr_name == "GeneratedProtocolMessageType":
+        return _parse_generated_protocol_message(rhs, module_globals, inner_fields, qualname)
+    if attr_name == "EnumTypeWrapper":
+        return _parse_enum_type_wrapper(rhs, module_globals, qualname)
+    return None
+
+
+def _parse_enum_value(node, value):
+    return {node.name: TypeClass(value.pytype())}
+
+
+def _parse_generated_protocol_message(call, module_globals, inner_fields, qualname):
     def parse_name(var):
         if not isinstance(var, astroid.Name):
             return None
         outer_node = module_globals.get(var.name)
         filtered_fields = inner_fields.get(var.name, {})
         return _parse_descriptor(outer_node, filtered_fields, qualname)
-    if not isinstance(node, astroid.AssignName):
-        return None
-    call = node.parent.value
     if not isinstance(call, astroid.Call) or len(call.args) < 3:
         return None
     type_dict = call.args[2]
@@ -345,6 +401,17 @@ def _extract_fields(node, module_globals, inner_fields, qualname):
     return None
 
 
+def _parse_enum_type_wrapper(rhs, module_globals, qualname):
+    def parse_name(var):
+        if not isinstance(var, astroid.Name):
+            return None
+        outer_node = module_globals.get(var.name)
+        filtered_fields = {}  # FIXME
+        return _parse_enum_descriptor(outer_node, filtered_fields, qualname)
+    enum_descriptor = rhs.args[0]
+    return parse_name(enum_descriptor)
+
+
 def _parse_message_type(node):
     try:
         outer = node.expr.value.expr.name
@@ -415,6 +482,9 @@ def import_(node, module_name, scope):
     def qualified_name(n):
         return '{}.{}'.format(module_name, n)
 
+    def unqualified_name(n):
+        return n[len(module_name)+1:]  # +1 = include dot
+
     new_names = {}
     for original_name, nodes in mod.globals.items():
         if likely_name(original_name):
@@ -427,6 +497,10 @@ def import_(node, module_name, scope):
 
     new_scope[module_name] = Module(module_name, new_names)
     for name, alias in imported_names:  # check aliasing for ImportFrom
+        if name == '*':
+            for qualname in new_names:
+                new_scope[unqualified_name(qualname)] = new_names[qualname]
+            break  # it's a SyntaxError to have other clauses with a *-import
         if alias is None:
             alias = name
         new_scope[alias] = new_names[qualified_name(name)]
